/*
 *    KEAFilter.java
 *    Copyright (C) 2000-2005 Eibe Frank, Olena Medelyan
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

import java.util.*;
import weka.core.*;
import weka.filters.*;
import weka.classifiers.*;

/**
 * This filter converts the incoming data into data appropriate for
 * keyphrase classification. It assumes that the dataset contains two
 * string attributes. The first attribute should contain the text of a
 * document. The second attribute should contain the keyphrases
 * associated with that document (if present). 
 *
 * The filter converts every instance (i.e. document) into a set of
 * instances, one for each word-based n-gram in the document. The
 * string attribute representing the document is replaced by some
 * numeric features, the estimated probability of each n-gram being a
 * keyphrase, and the rank of this phrase in the document according to
 * the probability.  Each new instances also has a class value
 * associated with it. The class is "true" if the n-gram is a true
 * keyphrase, and "false" otherwise. Of course, if the input document
 * doesn't come with author-assigned keyphrases, the class values for
 * that document will be missing.  
 *
 * @author Eibe Frank (eibe@cs.waikato.ac.nz), Olena Medelyan (olena@cs.waikato.ac.nz)
 * @version 2.0
 */
public class KEAFilter extends Filter implements OptionHandler {
	
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	/** Index of attribute containing the documents */
	private int m_DocumentAtt = 0;
	
	/** Index of attribute containing the keyphrases */
	private int m_KeyphrasesAtt = 1;
	
	/** The maximum length of phrases */
	private int m_MaxPhraseLength = 3;
	
	/** The minimum length of phrases */
	private int m_MinPhraseLength = 1;
	
	
	/** The number of phrases to extract. */
	private int m_numPhrases = 5;
	
	/** Is keyphrase frequency attribute being used? */
	private boolean m_KFused = false;
	
	
	// start: new features in KEA-4.0:    
	
	/** Should non-descriptors be replaced by corresponding descriptors? */
	private boolean m_DESCRreplace = true;
	
	/** If m_RELused = true, should the feature of their appearance/non-appearance be used? */
	private boolean m_APfeature = false;
	
	/** Is the node degree (number of related terms in candidate set) being used? */
	private boolean m_NODEfeature = true;
	
	/** Is the length of a phrase in words being used?*/
	private boolean m_LENGTHfeature = true;
	
	// end. Don't use these features with m_KFused or adjust indicies below.
	
	
	/** Flag for debugging mode */
	private boolean m_Debug = false;
	
	/** Determines whether internal periods are allowed */
	private boolean m_DisallowInternalPeriods = false;
	
	/** The minimum number of occurences of a phrase */
	private int m_MinNumOccur = 2;
	
	/** The number of features describing a phrase */
	private int m_NumFeatures = 2;
	
	/* Indices of attributes in m_ClassifierData */
	private int m_TfidfIndex = 0;
	private int m_FirstOccurIndex = 1;
	private int m_KeyFreqIndex = 2;
	
	/* Indicies of attributes for new features for KEA-4.0 */
	private int m_APIndex = 2;
	private int m_NodeIndex = 2;
	private int m_LengthIndex = 3;
	
	/** The punctuation filter used by this filter */
	private KEAPhraseFilter m_PunctFilter = null;
	
	/** The numbers filter used by this filter */
	private NumbersFilter m_NumbersFilter = null;
	
	/** The actual classifier used to compute probabilities */
	private DistributionClassifier m_Classifier = null;
	
	/** The dictionary containing the document frequencies */
	private HashMap m_Dictionary = null;
	
	/** The dictionary containing the keyphrases */
	private HashMap m_KeyphraseDictionary = null;
	
	/** The number of documents in the global frequencies corpus */
	private int m_NumDocs = 0;
	
	/** Template for the classifier data */
	private Instances m_ClassifierData = null;
	
	/** The default stemmer to be used */
	private Stemmer m_Stemmer = new SremovalStemmer();
	
	/** The list of stop words to be used */
	private Stopwords m_Stopwords = new StopwordsEnglish();
	
	/** The Vocabulary object */
	public static Vocabulary m_Vocabulary;
	
	private String m_vocabulary = "agrovoc";
	private String m_vocabularyFormat = "skos";
	
	/**
	 * Get the M_Vocabulary value.
	 * @return the M_Vocabulary value.
	 */
	public String getVocabulary() {	
		return m_vocabulary;
	}
	
	/**
	 * Set the M_Vocabulary value.
	 * @param newM_Vocabulary The new M_Vocabulary value.
	 */
	public void setVocabulary(String newM_Vocabulary) {	
		this.m_vocabulary = newM_Vocabulary;		
	}
	
	/**
	 * Get the M_VocabularyFormat value.
	 * @return the M_VocabularyFormat value.
	 */
	public String getVocabularyFormat() {	
		return m_vocabularyFormat;
	}
	
	/**
	 * Set the M_VocabularyFormat value.
	 * @param newM_VocabularyFormat The new M_VocabularyFormat value.
	 */
	public void setVocabularyFormat(String newM_VocabularyFormat) {	
		this.m_vocabularyFormat = newM_VocabularyFormat;		
	}
	
	
	/** Determines whether check for proper nouns is performed */
	private boolean m_CheckForProperNouns = true;
	
	/**
	 * Get the M_CheckProperNouns value.
	 * @return the M_CheckProperNouns value.
	 */
	public boolean getCheckForProperNouns() {
		return m_CheckForProperNouns;
	}
	
	/**
	 * Set the M_CheckProperNouns value.
	 * @param newM_CheckProperNouns The new M_CheckProperNouns value.
	 */
	public void setCheckForProperNouns(boolean newM_CheckProperNouns) {
		this.m_CheckForProperNouns = newM_CheckProperNouns;
	}
	
	/**
	 * Get the M_Stopwords value.
	 * @return the M_Stopwords value.
	 */
	public Stopwords getStopwords() {	
		return m_Stopwords;
	}
	
	/**
	 * Set the M_Stopwords value.
	 * @param newM_Stopwords The new M_Stopwords value.
	 */
	public void setStopwords(Stopwords newM_Stopwords) {	
		this.m_Stopwords = newM_Stopwords;		
	}
	
	
	/**
	 * Get the Stemmer value.
	 * @return the Stemmer value.
	 */
	public Stemmer getStemmer() {
		return m_Stemmer;
	}
	
	/**
	 * Set the Stemmer value.
	 * @param newStemmer The new Stemmer value.
	 */
	public void setStemmer(Stemmer newStemmer) {	
		this.m_Stemmer = newStemmer;
		
	}

	/**
	 * Get the value of MinNumOccur.
	 *
	 * @return Value of MinNumOccur.
	 */
	public int getMinNumOccur() {    
		return m_MinNumOccur;
	}
	
	/**
	 * Set the value of MinNumOccur.
	 *
	 * @param newMinNumOccur Value to assign to MinNumOccur.
	 */
	public void setMinNumOccur(int newMinNumOccur) {	
		m_MinNumOccur = newMinNumOccur;
	}
	
	/**
	 * Get the value of MaxPhraseLength.
	 *
	 * @return Value of MaxPhraseLength.
	 */
	public int getMaxPhraseLength() {	
		return m_MaxPhraseLength;
	}
	
	/**
	 * Set the value of MaxPhraseLength.
	 *
	 * @param newMaxPhraseLength Value to assign to MaxPhraseLength.
	 */
	public void setMaxPhraseLength(int newMaxPhraseLength) {	
		m_MaxPhraseLength = newMaxPhraseLength;
	}
	
	/**
	 * Get the value of MinPhraseLength.
	 *
	 * @return Value of MinPhraseLength.
	 */
	public int getMinPhraseLength() {	
		return m_MinPhraseLength;
	}
	
	/**
	 * Set the value of MinPhraseLength.
	 *
	 * @param newMinPhraseLength Value to assign to MinPhraseLength.
	 */
	public void setMinPhraseLength(int newMinPhraseLength) {	
		m_MinPhraseLength = newMinPhraseLength;
	}
	
	/**
	 * Get the value of numPhrases.
	 *
	 * @return Value of numPhrases.
	 */
	public int getNumPhrases() {    
		return m_numPhrases;
	}
	
	/**
	 * Set the value of numPhrases.
	 *
	 * @param newnumPhrases Value to assign to numPhrases.
	 */
	public void setNumPhrases(int newnumPhrases) {    
		m_numPhrases = newnumPhrases;
	}
	
	/**
	 * Returns the index of the stemmed phrases in the output ARFF file.
	 */
	public int getStemmedPhraseIndex() {	
		return m_DocumentAtt;
	}
	
	/**
	 * Returns the index of the unstemmed phrases in the output ARFF file.
	 */
	public int getUnstemmedPhraseIndex() {
		return m_DocumentAtt + 1;
	}
	
	/**
	 * Returns the index of the phrases' probabilities in the output ARFF file.
	 */
	public int getProbabilityIndex() {
		int index = m_DocumentAtt + 4;
		
		if (m_Debug) {
			if (m_KFused) {
				index++;
			}
		}
		if (m_APfeature) {
			index++;
		}
		if (m_NODEfeature) {
			index++;
		}
		if (m_LENGTHfeature) {
			index++;
		}      
		return index;
	}
	
	/**
	 * Returns the index of the phrases' ranks in the output ARFF file.
	 */
	public int getRankIndex() {
		return getProbabilityIndex() + 1;
	}
	
	/**
	 * Get the value of DocumentAtt.
	 *
	 * @return Value of DocumentAtt.
	 */
	public int getDocumentAtt() {
		return m_DocumentAtt;
	}
	
	/**
	 * Set the value of DocumentAtt.
	 *
	 * @param newDocumentAtt Value to assign to DocumentAtt.
	 */
	public void setDocumentAtt(int newDocumentAtt) {
		m_DocumentAtt = newDocumentAtt;
	}
	
	/**
	 * Get the value of KeyphraseAtt.
	 *
	 * @return Value of KeyphraseAtt.
	 */
	public int getKeyphrasesAtt() {	
		return m_KeyphrasesAtt;
	}
	
	/**
	 * Set the value of KeyphrasesAtt.
	 *
	 * @param newKeyphrasesAtt Value to assign to KeyphrasesAtt.
	 */
	public void setKeyphrasesAtt(int newKeyphrasesAtt) {
		m_KeyphrasesAtt = newKeyphrasesAtt;
	}
	
	
	/**
	 * Get the value of Debug.
	 *
	 * @return Value of Debug.
	 */
	public boolean getDebug() {
		return m_Debug;
	}
	
	/**
	 * Set the value of Debug.
	 *
	 * @param newDebug Value to assign to Debug.
	 */
	public void setDebug(boolean newDebug) {    
		m_Debug = newDebug;
	}
	
	/**
	 * Sets whether keyphrase frequency attribute is used.
	 */
	public void setKFused(boolean flag) {	
		m_KFused = flag;
		if (flag) {
			m_NumFeatures++;
		}
	}
	
	/**
	 * Sets whether Vocabulary relation attribute is used.
	 */
	public void setNumFeature() {
		if (m_APfeature) {
			m_NumFeatures++;
		}
		if (m_NODEfeature) {
			m_NumFeatures++;
		}
		if (m_LENGTHfeature) {
			m_NumFeatures++;
		}
	}
	
	/**
	 * Gets whether keyphrase frequency attribute is used.
	 */
	public boolean getKFused() {
		return m_KFused;
	}
	
	/**
	 * Get whether the supplied columns are to be processed
	 *
	 * @return true if the supplied columns won't be processed
	 */
	public boolean getDisallowInternalPeriods() {	
		return m_DisallowInternalPeriods;
	}
	
	/**
	 * Set whether selected columns should be processed. If true the 
	 * selected columns won't be processed.
	 *
	 * @param disallow the new invert setting
	 */
	public void setDisallowInternalPeriods(boolean disallow) {
		m_DisallowInternalPeriods = disallow;
	}
	
	
	public void loadThesaurus(Stemmer st, Stopwords sw) {
		m_Vocabulary = new Vocabulary(m_vocabulary,m_vocabularyFormat);
		m_Vocabulary.setStemmer(st);
		m_Vocabulary.setStopwords(sw);
		m_Vocabulary.initialize();
		try {
				
		if (m_DESCRreplace) {
			m_Vocabulary.buildUSE();
		}
		if (m_NODEfeature) {
			m_Vocabulary.buildREL();
		}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	
	/**
	 * Parses a given list of options controlling the behaviour of this object.
	 * Valid options are:<p>
	 *
	 * -K<br>
	 * Specifies whether keyphrase frequency statistic is used.<p>
	 *
	 * -R<br>
	 * Specifies whether Vocabulary relation statistic is used.<p>
	 *
	 * -M length<br>
	 * Sets the maximum phrase length (default: 3).<p>
	 *
	 * -L length<br>
	 * Sets the minimum phrase length (default: 1).<p>
	 *
	 * -D<br>
	 * Turns debugging mode on.<p>
	 *
	 * -I index<br>
	 * Sets the index of the attribute containing the documents (default: 0).<p>
	 *
	 * -J index<br>
	 * Sets the index of the attribute containing the keyphrases (default: 1).<p>
	 *
	 * -P<br>
	 * Disallow internal periods <p>
	 *
	 * -O number<br>
	 * The minimum number of times a phrase needs to occur (default: 2). <p>
	 *
	 * @param options the list of options as an array of strings
	 * @exception Exception if an option is not supported
	 */
	public void setOptions(String[] options) throws Exception {
		
		setKFused(Utils.getFlag('K', options));
		setDebug(Utils.getFlag('D', options));
		String docAttIndexString = Utils.getOption('I', options);
		if (docAttIndexString.length() > 0) {
			setDocumentAtt(Integer.parseInt(docAttIndexString) - 1);
		} else {
			setDocumentAtt(0);
		}
		String keyphraseAttIndexString = Utils.getOption('J', options);
		if (keyphraseAttIndexString.length() > 0) {
			setKeyphrasesAtt(Integer.parseInt(keyphraseAttIndexString) - 1);
		} else {
			setKeyphrasesAtt(1);
		}
		String maxPhraseLengthString = Utils.getOption('M', options);
		if (maxPhraseLengthString.length() > 0) {
			setMaxPhraseLength(Integer.parseInt(maxPhraseLengthString));
		} else {
			setMaxPhraseLength(3);
		}
		String minPhraseLengthString = Utils.getOption('M', options);
		if (minPhraseLengthString.length() > 0) {
			setMinPhraseLength(Integer.parseInt(minPhraseLengthString));
		} else {
			setMinPhraseLength(1);
		}
		String minNumOccurString = Utils.getOption('O', options);
		if (minNumOccurString.length() > 0) {
			setMinNumOccur(Integer.parseInt(minNumOccurString));
		} else {
			setMinNumOccur(2);
		}
		setDisallowInternalPeriods(Utils.getFlag('P', options));
	}
	
	/**
	 * Gets the current settings of the filter.
	 *
	 * @return an array of strings suitable for passing to setOptions
	 */
	public String [] getOptions() {
		
		String [] options = new String [13];
		int current = 0;
		
		if (getKFused()) {
			options[current++] = "-K";
		}
		if (getDebug()) {
			options[current++] = "-D";
		}
		options[current++] = "-I"; 
		options[current++] = "" + (getDocumentAtt() + 1);
		options[current++] = "-J"; 
		options[current++] = "" + (getKeyphrasesAtt() + 1);
		options[current++] = "-M"; 
		options[current++] = "" + (getMaxPhraseLength());
		options[current++] = "-L"; 
		options[current++] = "" + (getMinPhraseLength());
		options[current++] = "-O"; 
		options[current++] = "" + (getMinNumOccur());
		
		if (getDisallowInternalPeriods()) {
			options[current++] = "-P";
		}
		
		while (current < options.length) {
			options[current++] = "";
		}
		return options;
	}
	
	/**
	 * Returns an enumeration describing the available options
	 *
	 * @return an enumeration of all the available options
	 */
	public Enumeration listOptions() {
		
		Vector newVector = new Vector(7);
		
		newVector.addElement(new Option(
				"\tSpecifies whether keyphrase frequency statistic is used.",
				"K", 0, "-K"));
		newVector.addElement(new Option(
				"\tSets the maximum phrase length (default: 3).",
				"M", 1, "-M <length>"));
		newVector.addElement(new Option(
				"\tSets the minimum phrase length (default: 1).",
				"L", 1, "-L <length>"));
		newVector.addElement(new Option(
				"\tTurns debugging mode on.",
				"D", 0, "-D"));
		newVector.addElement(new Option(
				"\tSets the index of the document attribute (default: 0).",
				"I", 1, "-I"));
		newVector.addElement(new Option(
				"\tSets the index of the keyphrase attribute (default: 1).",
				"J", 1, "-J"));
		newVector.addElement(new Option(
				"\tDisallow internal periods.",
				"P", 0, "-P"));
		newVector.addElement(new Option(
				"\tSet the minimum number of occurences (default: 2).",
				"O", 1, "-O"));
		
		return newVector.elements();
	}
	
	/**
	 * Returns a string describing this filter
	 *
	 * @return a description of the filter suitable for
	 * displaying in the explorer/experimenter gui
	 */
	public String globalInfo() {
		return "Converts incoming data into data appropriate for " +
		"keyphrase classification.";
	}
	
	/**
	 * Sets the format of the input instances.
	 *
	 * @param instanceInfo an Instances object containing the input
	 * instance structure (any instances contained in the object are
	 * ignored - only the structure is required).
	 * @return true if the outputFormat may be collected immediately 
	 */
	public boolean setInputFormat(Instances instanceInfo) throws Exception {
		
		if (instanceInfo.classIndex() >= 0) {
			throw new Exception("Don't know what do to if class index set!");
		}
		if (!instanceInfo.attribute(m_KeyphrasesAtt).isString() ||
				!instanceInfo.attribute(m_DocumentAtt).isString()) {
			throw new Exception("Keyphrase attribute and document attribute " +
			"need to be string attributes.");
		}
		m_PunctFilter = new KEAPhraseFilter();
		int[] arr = new int[1];
		arr[0] = m_DocumentAtt;
		m_PunctFilter.setAttributeIndicesArray(arr);
		m_PunctFilter.setInputFormat(instanceInfo);
		m_PunctFilter.setDisallowInternalPeriods(getDisallowInternalPeriods());
		m_NumbersFilter = new NumbersFilter();
		m_NumbersFilter.setInputFormat(m_PunctFilter.getOutputFormat());
		super.setInputFormat(m_NumbersFilter.getOutputFormat());
		return false;
	}
	
	/**
	 * Input an instance for filtering. Ordinarily the instance is processed
	 * and made available for output immediately. Some filters require all
	 * instances be read before producing output.
	 *
	 * @param instance the input instance
	 * @return true if the filtered instance may now be
	 * collected with output().
	 * @exception Exception if the input instance was not of the correct 
	 * format or if there was a problem with the filtering.
	 */
	public boolean input(Instance instance) throws Exception {
		if (getInputFormat() == null) {
			throw new Exception("No input instance format defined");
		}
		if (m_NewBatch) {
			resetQueue();
			m_NewBatch = false;
		}
		
		if (m_Debug) {
			System.err.println("-- Reading instance");
		}
	
		m_PunctFilter.input(instance);
		m_PunctFilter.batchFinished();
		instance = m_PunctFilter.output();
		
		m_NumbersFilter.input(instance);
		m_NumbersFilter.batchFinished();
		instance = m_NumbersFilter.output();		
		
		if (m_Dictionary == null) {
			bufferInput(instance);
			return false;
		} else {
			FastVector vector = convertInstance(instance, false);
			Enumeration en = vector.elements();
			while (en.hasMoreElements()) {
				Instance inst = (Instance)en.nextElement();
				push(inst);
			}
			return true;
		}

	}
	
	/**
	 * Signify that this batch of input to the filter is finished. 
	 * If the filter requires all instances prior to filtering,
	 * output() may now be called to retrieve the filtered instances.
	 *
	 * @return true if there are instances pending output
	 * @exception Exception if no input structure has been defined
	 */
	public boolean batchFinished() throws Exception {
		
		if (getInputFormat() == null) {
			throw new Exception("No input instance format defined");
		}
		

		
		if (m_Dictionary == null) {
			buildGlobalDictionaries();
			buildClassifier();
			convertPendingInstances();
		} 
		flushInput();
		m_NewBatch = true;
		return (numPendingOutput() != 0);
	}
	
	
	/**
	 * Builds the global dictionaries.
	 */
	private void buildGlobalDictionaries() throws Exception {
		if (m_Debug) {
			System.err.println("--- Building global dictionaries");
		}
		
		// Build dictionary of n-grams with associated
		// document frequencies
		m_Dictionary = new HashMap();
		for (int i = 0; i < getInputFormat().numInstances(); i++) {
			String str = getInputFormat().instance(i).stringValue(m_DocumentAtt);
			HashMap hash = getPhrasesForDictionary(str);
			Iterator it = hash.keySet().iterator();
			while (it.hasNext()) {
				String phrase = (String)it.next();
				Counter counter = (Counter)m_Dictionary.get(phrase);
				if (counter == null) {
					m_Dictionary.put(phrase, new Counter());
				} else {
					counter.increment();
				}
			}
		}
		
		if (m_KFused) {       
			if (m_Debug) {
				System.out.println("KF_used feature");
			}
			
			// Build dictionary of n-grams that occur as keyphrases
			// with associated keyphrase frequencies
			m_KeyphraseDictionary = new HashMap();
			for (int i = 0; i < getInputFormat().numInstances(); i++) {
				String str = getInputFormat().instance(i).stringValue(m_KeyphrasesAtt);
				HashMap hash = getGivenKeyphrases(str, false);
				if (hash != null) {
					Iterator it = hash.keySet().iterator();
					while (it.hasNext()) {
						String phrase = (String)it.next();
						Counter counter = (Counter)m_KeyphraseDictionary.get(phrase);
						if (counter == null) {
							m_KeyphraseDictionary.put(phrase, new Counter());
						} else {
							counter.increment();
						}
					}
				}
			}
		} else {
			m_KeyphraseDictionary = null;
		}
		
		// Set the number of documents in the global corpus
		m_NumDocs = getInputFormat().numInstances();
	}
	
	/**
	 * Builds the classifier.
	 */
	// aly: The main function, where everything important happens
	private void buildClassifier() throws Exception {
		// Generate input format for classifier
		FastVector atts = new FastVector();
		for (int i = 0; i < getInputFormat().numAttributes(); i++) {
			if (i == m_DocumentAtt) {
				atts.addElement(new Attribute("TFxIDF"));
				atts.addElement(new Attribute("First_occurrence"));
				if (m_KFused) {
					atts.addElement(new Attribute("Keyphrase_frequency"));
				}
				if (m_APfeature) {
					FastVector rvals = new FastVector(2);
					rvals.addElement("False");
					rvals.addElement("True");
					atts.addElement(new Attribute("Vocabulary_relation",rvals));
				}
				if (m_NODEfeature) {
					atts.addElement(new Attribute("Relations_number"));
				}
				if (m_LENGTHfeature) {
					atts.addElement(new Attribute("Phrase_length"));
				}
			} else if (i == m_KeyphrasesAtt) {
				FastVector vals = new FastVector(2);
				vals.addElement("False");
				vals.addElement("True");
				atts.addElement(new Attribute("Keyphrase?", vals));
			} 
		}
		m_ClassifierData = new Instances("ClassifierData", atts, 0);
		m_ClassifierData.setClassIndex(m_NumFeatures);
		
		if (m_Debug) {
			System.err.println("--- Converting instances for classifier");
		}
		
		// Convert pending input instances into data for classifier
		for(int i = 0; i < getInputFormat().numInstances(); i++) {
			Instance current = getInputFormat().instance(i);
			
			// Get the key phrases for the document
			String keyphrases = current.stringValue(m_KeyphrasesAtt);
			HashMap hashKeyphrases = getGivenKeyphrases(keyphrases, false);
			HashMap hashKeysEval = getGivenKeyphrases(keyphrases, true);
			
			// Get the phrases for the document
			HashMap hash = new HashMap();
			int length = getPhrases(hash, current.stringValue(m_DocumentAtt));
			
			// Compute the feature values for each phrase and
			// add the instance to the data for the classifier
			
			Iterator it = hash.keySet().iterator();
			while (it.hasNext()) {
				String phrase = (String)it.next();
				FastVector phraseInfo = (FastVector)hash.get(phrase);
				
				double[] vals =  featVals(phrase, phraseInfo, true,
						hashKeysEval, hashKeyphrases, length, hash);
				Instance inst = new Instance(current.weight(), vals);
				m_ClassifierData.add(inst);
			}
		}
		
		if (m_Debug) {
			System.err.println("--- Building classifier");
		}
		
		// Build classifier
		FilteredClassifier fclass = new FilteredClassifier();
		fclass.setClassifier(new weka.classifiers.NaiveBayesSimple());
		fclass.setFilter(new DiscretizeFilter());
		m_Classifier = fclass;
		// System.out.print(m_ClassifierData);    
		m_Classifier.buildClassifier(m_ClassifierData);
		
		if (m_Debug) {
			System.err.println(m_Classifier);
		}
		
		// Save space
		m_ClassifierData = new Instances(m_ClassifierData, 0);
	}
	
	/** 
	 * Conmputes the feature values for a given phrase.
	 */
	private double[] featVals(String id, FastVector phraseInfo, 
			boolean training, HashMap hashKeysEval,
			HashMap hashKeyphrases, int length, HashMap hash) {
		
		// Compute feature values
		Counter counterLocal = (Counter)phraseInfo.elementAt(1);
		double[] newInst = new double[m_NumFeatures + 1];
		
		
		// Compute TFxIDF
		Counter counterGlobal = (Counter)m_Dictionary.get(id);
		double localVal = counterLocal.value(), globalVal = 0;
		if (counterGlobal != null) {
			globalVal = counterGlobal.value();
			if (training) {
				globalVal = globalVal - 1;
			}
		}
		
		// Just devide by length to get approximation of probability
		// that phrase in document is our phrase
		newInst[m_TfidfIndex] = (localVal / ((double)length)) *
		(-Math.log((globalVal + 1)/ ((double)m_NumDocs + 1)));
		
		// Compute first occurrence
		Counter counterFirst = (Counter)phraseInfo.elementAt(0);
		newInst[m_FirstOccurIndex] = (double)counterFirst.value() /
		(double)length;
		
		// Is keyphrase frequency attribute being used?
		if (m_KFused) {
			Counter keyphraseC = (Counter)m_KeyphraseDictionary.get(id);
			if ((training) && (hashKeyphrases != null) &&
					(hashKeyphrases.containsKey(id))) {
				newInst[m_KeyFreqIndex] = keyphraseC.value() - 1;
			} else {
				if (keyphraseC != null) {
					newInst[m_KeyFreqIndex] = keyphraseC.value();
				} else {
					newInst[m_KeyFreqIndex] = 0;
				}
			}
		}
		
		// Is term appearance attribute being used?
		if (m_APfeature) {	
			String relation = (String)phraseInfo.elementAt(3);
			if (relation.equals("False")) {
				newInst[m_APIndex] = 0.0;
			} else {
				newInst[m_APIndex] = 1.0;
			}
		} 
		
		// Is node degree attribute being used?   
		if (m_NODEfeature) {	
			
			Vector idsRT = (Vector)m_Vocabulary.getRelated(id);
			
			int intern = 0;			
			if (idsRT != null) {
				for (int d = 0; d < idsRT.size(); d++) {
					//System.out.println(idsRT.elementAt(d));
					if (hash.get(idsRT.elementAt(d)) != null) {
						//System.out.println("\tin the document");
						intern++;
					}					
				}
			}
			// System.out.println(id + " has node degree " + intern);
			newInst[m_NodeIndex] = (double)intern;
			
		}
		
		// Is term length attribute being used?
		if (m_LENGTHfeature) {
			String original = (String)phraseInfo.elementAt(2);
			String [] words = split(original," ");
			newInst[m_LengthIndex] = (double)words.length;
		}
		
		// Compute class value
		
		if (hashKeysEval == null) { // no author-assigned keyphrases
			newInst[m_NumFeatures] = Instance.missingValue();
		} else if (!hashKeysEval.containsKey(id)) {
			// System.out.println(" - FALSE");
			newInst[m_NumFeatures] = 0; // No keyphrase
		} else {
			hashKeysEval.remove(id);
			newInst[m_NumFeatures] = 1; // Keyphrase
		}
		return newInst;
	}
	
	/**
	 * Sets output format and converts pending input instances.
	 */
	private void convertPendingInstances() throws Exception {
		
		if (m_Debug) {
			System.err.println("--- Converting pending instances");
		}
		
		// Create output format for filter
		FastVector atts = new FastVector();
		for (int i = 0; i < getInputFormat().numAttributes(); i++) {
			if (i == m_DocumentAtt) {
				// string attributes
				atts.addElement(new Attribute("N-gram", null));
				atts.addElement(new Attribute("N-gram-original", null));
				// numeric attributes
				atts.addElement(new Attribute("TFxIDF"));
				atts.addElement(new Attribute("First_occurrence"));
				// optional attributes
				if (m_Debug) {
					if (m_KFused) {
						atts.addElement(new Attribute("Keyphrase_frequency"));
					}
				}
				if (m_APfeature) {
					FastVector rvals = new FastVector(2);
					rvals.addElement("False");
					rvals.addElement("True");
					atts.addElement(new Attribute("Vocabulary_relation",rvals));
				}
				if (m_NODEfeature) {
					atts.addElement(new Attribute("Relations_number"));
				}
				if (m_LENGTHfeature) {
					atts.addElement(new Attribute("Phrase_length"));
				}
				
				atts.addElement(new Attribute("Probability"));
				atts.addElement(new Attribute("Rank"));
			} else if (i == m_KeyphrasesAtt) {
				FastVector vals = new FastVector(2);
				vals.addElement("False");
				vals.addElement("True");
				atts.addElement(new Attribute("Keyphrase?", vals));
			} else {
				atts.addElement(getInputFormat().attribute(i));
			}
		}
		Instances outFormat = new Instances("KEAdata", atts, 0);
		setOutputFormat(outFormat);
		
		// Convert pending input instances into output data
		for(int i = 0; i < getInputFormat().numInstances(); i++) {
			Instance current = getInputFormat().instance(i);
			FastVector vector = convertInstance(current, true);
			Enumeration en = vector.elements();
			while (en.hasMoreElements()) {
				Instance inst = (Instance)en.nextElement();
				push(inst);
			}
		}
	} 
	
	/**
	 * Converts an instance.
	 */
	private FastVector convertInstance(Instance instance, boolean training) 
	throws Exception {
		
		FastVector vector = new FastVector();
		
		if (m_Debug) {
			System.err.println("-- Converting instance");
		}
		
		// Get the key phrases for the document
		HashMap hashKeyphrases = null;
		HashMap hashKeysEval = null;
		if (!instance.isMissing(m_KeyphrasesAtt)) {
			String keyphrases = instance.stringValue(m_KeyphrasesAtt);
			hashKeyphrases = getGivenKeyphrases(keyphrases, false);
			hashKeysEval = getGivenKeyphrases(keyphrases, true);
		}
		
		// Get the phrases for the document
		HashMap hash = new HashMap();
		int length = getPhrases(hash, instance.stringValue(m_DocumentAtt));
		
		// Compute number of extra attributes
		int numFeatures = 5;
		if (m_Debug) {
			if (m_KFused) {
				numFeatures = numFeatures + 1;
			}
		} 
		if (m_APfeature) {
			numFeatures = numFeatures + 1;
		}
		if (m_NODEfeature) {	
			numFeatures = numFeatures + 1;
		}
		if (m_LENGTHfeature) {
			numFeatures = numFeatures + 1;
		}
		
		// Set indices of key attributes
		//int phraseAttIndex = m_DocumentAtt;
		int tfidfAttIndex = m_DocumentAtt + 2;
		int distAttIndex = m_DocumentAtt + 3;
		int probsAttIndex = m_DocumentAtt + numFeatures - 1;    
		//int classAttIndex = numFeatures;
		
		// Go through the phrases and convert them into instances
		Iterator it = hash.keySet().iterator();
		while (it.hasNext()) {
			String phrase = (String)it.next();
			FastVector phraseInfo = (FastVector)hash.get(phrase);
			
			double[] vals =  featVals(phrase, phraseInfo, training,
					hashKeysEval, hashKeyphrases, length, hash);
			
			Instance inst = new Instance(instance.weight(), vals);
			
			inst.setDataset(m_ClassifierData);
			
			// Get probability of phrase being key phrase
			double[] probs = m_Classifier.distributionForInstance(inst);
			double prob = probs[1];
			
			// Compute attribute values for final instance
			double[] newInst = 
				new double[instance.numAttributes() + numFeatures];
			int pos = 0;
			for (int i = 0; i < instance.numAttributes(); i++) {
				if (i == m_DocumentAtt) {
					
					// output of values for a given phrase:
					
					// Add phrase
					int index = outputFormatPeek().attribute(pos).
					addStringValue(phrase);
					newInst[pos++] = index;
					
					// Add original version
					index = outputFormatPeek().attribute(pos).
					addStringValue((String)phraseInfo.elementAt(2));
					newInst[pos++] = index;
					
					// Add TFxIDF
					newInst[pos++] = inst.value(m_TfidfIndex);
					
					// Add distance
					newInst[pos++] = inst.value(m_FirstOccurIndex);
					
					// Add other features
					if (m_Debug) {
						if (m_KFused) {
							newInst[pos++] = inst.value(m_KeyFreqIndex);
						}
					}
					if (m_APfeature) {
						newInst[pos++] = inst.value(m_APIndex);
					}
					if (m_NODEfeature) {
						newInst[pos++] = inst.value(m_NodeIndex);
					}
					if (m_LENGTHfeature) {
						newInst[pos++] = inst.value(m_LengthIndex);
					}
					
					// Add probability 
					probsAttIndex = pos;
					newInst[pos++] = prob;
					
					// Set rank to missing (computed below)
					newInst[pos++] = Instance.missingValue();
					
				} else if (i == m_KeyphrasesAtt) {
					newInst[pos++] = inst.classValue();
				} else {
					newInst[pos++] = instance.value(i);
				}
			}
			Instance ins = new Instance(instance.weight(), newInst);
			ins.setDataset(outputFormatPeek());
			vector.addElement(ins);
		}
		
		
		// Add dummy instances for keyphrases that don't occur
		// in the document
		if (hashKeysEval != null) {
			Iterator phrases = hashKeysEval.keySet().iterator();
			while (phrases.hasNext()) {
				String phrase = (String)phrases.next();
				
					
					double[] newInst = 
						new double[instance.numAttributes() + numFeatures];
					int pos = 0;
					for (int i = 0; i < instance.numAttributes(); i++) {
						if (i == m_DocumentAtt) {
							// System.out.println("Here: " + phrase);
							// Add phrase
							int index = outputFormatPeek().attribute(pos).
							addStringValue(phrase);
							newInst[pos++] = (double)index;
							
							// Add original version
							index = outputFormatPeek().attribute(pos).
							addStringValue((String)hashKeysEval.get(phrase));
							newInst[pos++] = (double)index;
							
							// Add TFxIDF
							newInst[pos++] = Instance.missingValue();
							
							// Add distance
							newInst[pos++] = Instance.missingValue();
							
							// Add other features
							if (m_Debug) {
								if (m_KFused) {
									newInst[pos++] = Instance.missingValue();
								}
							}
							if (m_APfeature) {
								newInst[pos++] = Instance.missingValue();
							}
							if (m_NODEfeature) {
								newInst[pos++] = Instance.missingValue();
							}
							if (m_LENGTHfeature) {
								newInst[pos++] = Instance.missingValue();
							}
							
							// Add probability and rank
							newInst[pos++] = -Double.MAX_VALUE;
							// newInst[pos++] = Instance.missingValue();
						} else if (i == m_KeyphrasesAtt) {
							newInst[pos++] = 1; // Keyphrase
						} else {
							newInst[pos++] = instance.value(i);
						} 
					
						Instance inst = new Instance(instance.weight(), newInst);
						inst.setDataset(outputFormatPeek());
						vector.addElement(inst);
					}
				
			}
		}
		
		// Sort phrases according to their distance (stable sort)
		double[] vals = new double[vector.size()];
		for (int i = 0; i < vals.length; i++) {
			vals[i] = ((Instance)vector.elementAt(i)).value(distAttIndex);
		}
		FastVector newVector = new FastVector(vector.size());
		int[] sortedIndices = Utils.stableSort(vals);
		for (int i = 0; i < vals.length; i++) {
			newVector.addElement(vector.elementAt(sortedIndices[i]));
		}
		vector = newVector;
		
		// Sort phrases according to their tfxidf value (stable sort)
		for (int i = 0; i < vals.length; i++) {
			vals[i] = -((Instance)vector.elementAt(i)).value(tfidfAttIndex);
		}
		newVector = new FastVector(vector.size());
		sortedIndices = Utils.stableSort(vals);
		for (int i = 0; i < vals.length; i++) {
			newVector.addElement(vector.elementAt(sortedIndices[i]));
		}
		vector = newVector;
		
		// Sort phrases according to their probability (stable sort)
		for (int i = 0; i < vals.length; i++) {
			vals[i] = 1 - ((Instance)vector.elementAt(i)).value(probsAttIndex);
		}
		newVector = new FastVector(vector.size());
		sortedIndices = Utils.stableSort(vals);
		for (int i = 0; i < vals.length; i++) {
			newVector.addElement(vector.elementAt(sortedIndices[i]));
		}
		vector = newVector;
		
		// Compute rank of phrases. Check for subphrases that are ranked
		// lower than superphrases and assign probability -1 and set the
		// rank to Integer.MAX_VALUE
		int rank = 1;
		for (int i = 0; i < vals.length; i++) {
			Instance currentInstance = (Instance)vector.elementAt(i);
			// Short cut: if phrase very unlikely make rank very low and continue
			if (Utils.grOrEq(vals[i], 1.0)) {
				currentInstance.setValue(probsAttIndex + 1, Integer.MAX_VALUE);
				continue;
			}
			
			// Otherwise look for super phrase starting with first phrase
			// in list that has same probability, TFxIDF value, and distance as
			// current phrase. We do this to catch all superphrases
			// that have same probability, TFxIDF value and distance as current phrase.
			int startInd = i;
			while (startInd < vals.length) {
				Instance inst = (Instance)vector.elementAt(startInd);
				if ((inst.value(tfidfAttIndex) != 
					currentInstance.value(tfidfAttIndex)) ||
					(inst.value(probsAttIndex) != 
						currentInstance.value(probsAttIndex)) ||
						(inst.value(distAttIndex) !=
							currentInstance.value(distAttIndex))) {
					break;
				}
				startInd++;
			}
			currentInstance.setValue(probsAttIndex + 1, rank++);
			
		}
		return vector;
	}
	
	

	
	/**
	 * Returns a hashtable. Fills the hashtable
	 * with the stemmed n-grams occuring in the given string
	 * (as keys) and the number of times it occurs.
	 */
	private HashMap getPhrasesForDictionary(String str) {
		
		String[] buffer = new String[m_MaxPhraseLength];
		HashMap hash = new HashMap();
		
		StringTokenizer tok = new StringTokenizer(str, "\n");
		while (tok.hasMoreTokens()) {
			String phrase = tok.nextToken();
			int numSeen = 0;
			StringTokenizer wordTok = new StringTokenizer(phrase, " ");
			while (wordTok.hasMoreTokens()) {
				String word = wordTok.nextToken();
				
				// Store word in buffer
				for (int i = 0; i < m_MaxPhraseLength - 1; i++) {
					buffer[i] = buffer[i + 1];
				}
				buffer[m_MaxPhraseLength - 1] = word;
				
				// How many are buffered?
				numSeen++;
				if (numSeen > m_MaxPhraseLength) {
					numSeen = m_MaxPhraseLength;
				}
				
				// Don't consider phrases that end with a stop word
				if (m_Stopwords.isStopword(buffer[m_MaxPhraseLength - 1])) {
					continue;
				}
				
				// Loop through buffer and add phrases to hashtable
				StringBuffer phraseBuffer = new StringBuffer();
				for (int i = 1; i <= numSeen; i++) {
					if (i > 1) {
						phraseBuffer.insert(0, ' ');
					}
					phraseBuffer.insert(0, buffer[m_MaxPhraseLength - i]);
					
					// Don't consider phrases that begin with a stop word
					if ((i > 1) && 
							(m_Stopwords.isStopword(buffer[m_MaxPhraseLength - i]))) {
						continue;
					}
					
					// Only consider phrases with minimum length
					if (i >= m_MinPhraseLength) {
						
						// Stem string
						String orig = phraseBuffer.toString();
						
						String id = (String)m_Vocabulary.getID(orig);
						
						// and if they appear in the Vocabulary
						if (id != null) { 
							
							// if non-descriptor should be replaced by descriptors
							if (m_DESCRreplace) {
													
								// get the id of the corresponding descriptor
								String iduse = (String)m_Vocabulary.getDescriptor(id);
								
								// if there is a descriptor, replace it:
								if (iduse != null) {
									id = iduse;
								}
							}
								Counter count = (Counter)hash.get(id);
								if (count == null) {
									hash.put(id, new Counter());
								} else {
									count.increment();
								}
							}
						} 
					}
				}
			}
			return hash;
		}
		
		
		/**
		 * Expects an empty hashtable. Fills the hashtable
		 * with the stemmed n-grams occuring in the given string
		 * (as keys). Stores the position, the number of occurences,
		 * and the most commonly occurring orgininal version of
		 * each n-gram.
		 *
		 * N-grams that occur less than m_MinNumOccur are not used.
		 *
		 * Returns the total number of words (!) in the string.
		 */
		private int getPhrases(HashMap hash, String str) {
			
			// hash = table to store all the information about phrases extracted from "str"
			// str  = the content of the document, separated by newlines in sentences
			
			String[] buffer = new String[m_MaxPhraseLength];
			
			// Extracting strings of a predefined length from "str":
			
			StringTokenizer tok = new StringTokenizer(str, "\n");
			int pos = 1; 
			
			while (tok.hasMoreTokens()) {
				String phrase = tok.nextToken();
				
				int numSeen = 0;
				StringTokenizer wordTok = new StringTokenizer(phrase, " ");
				while (wordTok.hasMoreTokens()) {
					String word = wordTok.nextToken();
					
					// Store word in buffer
					for (int i = 0; i < m_MaxPhraseLength - 1; i++) {
						buffer[i] = buffer[i + 1];
					}
					buffer[m_MaxPhraseLength - 1] = word;
					
					// How many are buffered?
					numSeen++;
					if (numSeen > m_MaxPhraseLength) {
						numSeen = m_MaxPhraseLength;
					}
					
					// Don't consider phrases that end with a stop word
					if (m_Stopwords.isStopword(buffer[m_MaxPhraseLength - 1])) {
						pos++;
						continue;
					}	
					
					// Loop through buffer and add phrases to hashtable
					StringBuffer phraseBuffer = new StringBuffer();
					for (int i = 1; i <= numSeen; i++) {
						if (i > 1) {
							phraseBuffer.insert(0, ' ');
						}
						phraseBuffer.insert(0, buffer[m_MaxPhraseLength - i]);
						
						// Don't consider phrases that begin with a stop word
						if ((i > 1) && 
								(m_Stopwords.isStopword(buffer[m_MaxPhraseLength - i]))) {
							continue;
						}
						
						// Final restriction:
						// Only consider phrases with minimum length
						if (i >= m_MinPhraseLength) {
							
							// phrStr = each detected phase  
							String phrStr = phraseBuffer.toString();
							
							// Create internal representation:
							// either a stemmed version or a pseudo phrase: 
							// String pseudo = pseudoPhrase(phrStr);
							
							// Match against the Vocabulary
							String id = (String)m_Vocabulary.getID(phrStr);
							
							
							// If its an Vocabulary phrase:
							if (id != null) {		
								String orig = m_Vocabulary.getOrig(id);
								
								//System.out.println(phrStr + " => " + orig);
								
								// if the phrase is a non-descriptor
								// and the DESCRreplace option is selected,
								// replace it with the corresponding descriptor
								// if it is available
								
								if (m_DESCRreplace) {
									String iduse = (String)m_Vocabulary.getDescriptor(id);
									if (iduse != null) {
										id = iduse;
									}	
								}
								
								// Get the vector of the current phrase from the hash table.
								// If it was already extracted from "str", the values will be
								// updated in next steps, if not a new vector will be created.
								
								FastVector vec = (FastVector)hash.get(id);
								
								if (vec == null) {
									
									// Specifying the size of the vector
									// According to additional selected features:
									
									if (m_APfeature) {
										vec = new FastVector(3);
									} else {
										vec = new FastVector(2);
									}
									
									// Update hashtable with all the info
									vec.addElement(new Counter(pos + 1 - i)); //0
									vec.addElement(new Counter()); //1
									vec.addElement(orig); //2
									if (m_APfeature) {
										vec.addElement("True"); //3	
									}
									hash.put(id, vec);
								} else {
									
									// If the phrase already was identified,
									// update its values in the old vector
									
									// Update number of occurrences
									((Counter)((FastVector)vec).elementAt(1)).increment();
									
									if (m_APfeature) {
										
										// If the given phrase was added to the list of candidates
										// before, because of it's relation to another term.
										// Set the relation value to Vocabulary Term (VT)
										// since it has more preference
										
										String before = (String)vec.elementAt(3);
										if (!(before.equals("False"))) {
											vec.setElementAt("True",3);
										}
									}
									
								}								
							} else {
								//System.out.println(phrStr);
							}
						}
					}
					pos++;
				}
			}
			
			// Replace secondary hashtables with most commonly occurring
			// version of each phrase (canonical) form. Delete all words
			// that are proper nouns.
			Iterator phrases = hash.keySet().iterator();
			
			while (phrases.hasNext()) {
				String phrase = (String)phrases.next();
				FastVector info = (FastVector)hash.get(phrase);
				
				// Occurring less than m_MinNumOccur?
				
				if (((Counter)((FastVector)info).elementAt(1)).value() < m_MinNumOccur) {
					phrases.remove();
					continue;
				}
			}
			return pos;
		}
		
		
		/** 
		 * Splits a string at given character into an array (ALY)
		 */
		private static String[] split(String str,String separator) {
			
			ArrayList lst = new ArrayList();
			String word = ""; 
			
			for (int i = 0; i < str.length(); i++) {
				int j = i + 1;
				String letter = str.substring(i,j);
				if (!letter.equalsIgnoreCase(separator)) {
					word = word + str.charAt(i);
				} else {
					lst.add(word);
					word = ""; 
				}
			}
			if (word != "") {
				lst.add(word);
			}
			String[] result = (String[])lst.toArray(new String[lst.size()]);
			return result;
		}	
		
		/**
		 * Gets all the phrases in the given string and puts them into the
		 * hashtable.  Also stores the original version of the stemmed
		 * phrase in the hash table.  
		 */
		private HashMap getGivenKeyphrases(String str,
				boolean forEval) {
			
			HashMap hash = new HashMap();
			
			StringTokenizer tok = new StringTokenizer(str, "\n");
			while (tok.hasMoreTokens()) {
				String orig = tok.nextToken();
				orig = orig.trim();
				
				if (orig.length() > 0) {					
					
					String id = (String)m_Vocabulary.getID(orig);					
					if (id != null) {
						if (!hash.containsKey(id)) {
							hash.put(id, orig);
							// System.out.println(orig + " was mapped to " + (String)m_Vocabulary.getOrig(id));
						} else {	
							// System.out.println(orig + " was not found!");
							if (forEval && m_Debug) {
								System.err.println("Skipping the phrase " + orig + ", which appears twice in the author-assigned keyphrase set.");
							}
						}
					}
				}
			}
			if (hash.size() == 0) {
				return null;
			} else {
				return hash;
			}
		}
		
		/**
		 * Main method for testing this class.
		 *
		 * @param argv should contain arguments to the filter: use -h for help
		 */
		public static void main(String [] argv) {
			
			try {
				if (Utils.getFlag('b', argv)) {
					Filter.batchFilterFile(new KEAFilter(), argv);
				} else {
					Filter.filterFile(new KEAFilter(), argv);
				}
			} catch (Exception ex) {
				System.err.println(ex.getMessage());
			}
		}
	}
	
	
	
	
	
	
	
	
	
	
	
